use crate::StdError;
use chrono::Utc;
use serde::Deserialize;
use serde_dhall::{SimpleType, StaticType};
use std::collections::HashMap;
use std::error::Error;
use std::fmt;
use std::path::Path;

#[derive(Deserialize, StaticType, Debug)]
pub enum Action {
    ActivateLayer(String),                      // Activate a layer
    Print(String),                              // Print a string to console
    MQTTPub { topic: String, payload: String }, // Publish payload on topic via MQTT
}

impl fmt::Display for Action {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match &self {
            Action::ActivateLayer(layer) => write!(f, "ActivateLayer({})", layer),
            Action::Print(p) => write!(f, "Print({})", p),
            Action::MQTTPub { topic, payload } => {
                write!(f, "MQTTPub(topic={},payload={})", topic, payload)
            }
        }
    }
}

// Newtype wrapper around a Vec of actions.
// serde(transparent) and the imp StaticType below ensures we ignore the outer struct and
// transparently deserialize into the inner vec.
#[derive(Deserialize, Debug)]
#[serde(transparent)]
pub struct Actions(Vec<Action>);

impl Actions {
    // Return the iterator over the inner Vec of Actions
    pub fn iter(&self) -> std::slice::Iter<Action> {
        self.0.iter()
    }
}

impl fmt::Display for Actions {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Actions(count={})", self.0.len())
    }
}

impl StaticType for Actions {
    fn static_type() -> SimpleType {
        SimpleType::List(Box::new(Action::static_type()))
    }
}

#[derive(Deserialize, StaticType, Debug)]
pub enum Mapping {
    NOP,                   // Do nothing.
    Passthrough,           // Passthrough to the layer below
    Trigger(Action),       // Trigger an action
    TriggerMulti(Actions), // Trigger multiple actions in sequence
}

impl fmt::Display for Mapping {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "{}",
            match &self {
                Mapping::NOP => "NOP".to_owned(),
                Mapping::Passthrough => "Passthrough".to_owned(),
                Mapping::Trigger(act) => format!("Trigger({})", act),
                Mapping::TriggerMulti(acts) => format!("Trigger({})", acts),
            },
        )
    }
}

#[derive(Deserialize, Debug)]
pub struct Layer {
    pub default_mapping: Mapping, // What do we do if there's no mapping for a button?
    pub mappings: HashMap<String, Mapping>, // Button id -> mapping
}

#[derive(Deserialize, Debug)]
pub struct MQTTServer {
    pub host: String,
    pub port: Option<u16>,
    pub user: Option<String>,
    pub pass: Option<String>,
    pub client_id: Option<String>,
    pub topic_prefix: Option<String>,
}

#[derive(Deserialize, Debug)]
pub struct Config {
    pub mqtt_servers: HashMap<String, MQTTServer>,
    pub layers: HashMap<String, Layer>,
}

impl Config {
    pub fn from_file(p: &Path) -> Result<Self, Box<dyn Error>> {
        serde_dhall::from_file(p)
            .parse::<Self>()
            .map_err(|e| e.into())
    }

    pub fn generate_prelude() -> StdError<()> {
        let mut m = HashMap::new();
        m.insert("Action".to_string(), Action::static_type());
        m.insert("Mapping".to_string(), Mapping::static_type());
        println!(
            "-- Sampad Dhall types, autogenerated by git#{} on {}",
            option_env!("VERSION").unwrap_or("unknown"),
            Utc::now(),
        );
        println!(
            "{{ Action = {}, Mapping = {} }}",
            Action::static_type(),
            Mapping::static_type()
        );
        Ok(())
    }
}
