use bitvec::prelude::*;
use hidapi::{DeviceInfo, HidApi, HidDevice};
use std::collections::VecDeque;
use std::error::Error;
use std::fmt;
use tracing::{event, Level};

const HID_USAGE_PAGE: u16 = 0xFF;
const HID_USAGE: u16 = 0x1;

pub fn get_devices<'a>(hid: &'a HidApi, print: bool) -> Vec<&'a DeviceInfo> {
    if print {
        println!("All HIDs attached:");
        println!("Vendor ID, Product ID, Manufacturer, Produce, Usage Page, Usage");
    }
    let mut devices = vec![];
    for device in hid.device_list() {
        if (device.usage_page(), device.usage()) == (HID_USAGE_PAGE, HID_USAGE) {
            devices.push(device);
        }
        if print {
            println!(
                "{}, {}, {}, {}, {}, {}",
                device.vendor_id(),
                device.product_id(),
                device.manufacturer_string().unwrap_or("Unknown"),
                device.product_string().unwrap_or("Unknown"),
                device.usage_page(),
                device.usage(),
            );
        }
    }
    devices
}

#[derive(Debug)]
pub enum ButtonEvent {
    KeyDown(u8),
    KeyUp(u8),
}

impl fmt::Display for ButtonEvent {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let vals = match *self {
            Self::KeyDown(key) => ("KeyDown", key),
            Self::KeyUp(key) => ("KeyUp", key),
        };
        write!(f, "{}(key={})", vals.0, vals.1)
    }
}

type State = u16;

pub struct ButtonPad<'a> {
    dev: &'a HidDevice,
    state: State,
    queued: VecDeque<ButtonEvent>,
}

impl ButtonPad<'_> {
    pub fn new(dev: &HidDevice) -> ButtonPad {
        ButtonPad {
            dev,
            state: 0,
            queued: VecDeque::new(),
        }
    }
}

impl Iterator for ButtonPad<'_> {
    type Item = Result<ButtonEvent, Box<dyn Error>>;

    fn next(&mut self) -> Option<Self::Item> {
        loop {
            // If we have queued events pop them first
            if let Some(ev) = self.queued.pop_front() {
                return Some(Ok(ev));
            }
            // Read from the device, blocking until the next report
            let mut buf: [u8; 2] = [0; 2];
            self.dev.read(&mut buf).expect("error reading");
            event!(Level::INFO, ?buf, "Read from device");

            let new: u16 = buf[0] as u16 | (buf[1] as u16) << 8;

            // Work out what, if anything, changed between the last report and this one, and
            // queue all the
            self.queued.extend(
                self.state
                    .view_bits::<Lsb0>()
                    .iter()
                    .zip(new.view_bits::<Lsb0>().iter())
                    .enumerate()
                    .filter_map(
                        |(i, (state_last, state_new))| match (*state_last, *state_new) {
                            (false, true) => Some(ButtonEvent::KeyDown(i.try_into().unwrap())),
                            (true, false) => Some(ButtonEvent::KeyUp(i.try_into().unwrap())),
                            _ => None,
                        },
                    ),
            );
            self.state = new;
        }
    }
}
